#include "runtime/renderer_api.h"
#include "runtime/config.h"
#include "runtime/handles.h"
#include "runtime/threading.h"

#include "gpu.h"
#include "pipelines.h"

#include <volk/volk.h>
#include <stdlib.h>

RT_CVAR_I(r_VkMaxPipelineCount, "Maximum number of pipeline objects. Default: 1024", 1024);

typedef struct rt_pipeline_s {
    uint32_t version;
    rt_pipeline pipeline;
    struct rt_pipeline_s *next_free;
} rt_pipeline_slot;

static rt_pipeline_slot *_pipelines;
static rt_pipeline_slot *_first_free;
static rt_rwlock _lock;

rt_result InitPipelineManagement(void) {
    rt_create_rwlock_result lock_res = rtCreateRWLock();
    if (!lock_res.ok)
        return RT_UNKNOWN_ERROR;
    _lock = lock_res.lock;

    _pipelines = calloc(r_VkMaxPipelineCount.i, sizeof(rt_pipeline_slot));
    if (!_pipelines) {
        rtDestroyRWLock(&_lock);
        return RT_OUT_OF_MEMORY;
    }
    /* Keep [0] unused to preserve 0 as the invalid handle */
    _first_free = &_pipelines[1];
    for (int i = 1; i < r_VkMaxPipelineCount.i - 1; ++i) {
        _pipelines[i].next_free = &_pipelines[i + 1];
    }
    return RT_SUCCESS;
}

static void DestroyPipeline(rt_pipeline_slot *slot) {
    if (slot->pipeline.pipeline) {
        vkDestroyPipeline(g_gpu.device, slot->pipeline.pipeline, g_gpu.alloc_cb);
    }
    slot->next_free = _first_free;
    _first_free     = slot;
}

void ShutdownPipelineManagement(void) {
    for (int i = 1; i < r_VkMaxPipelineCount.i; ++i) {
        DestroyPipeline(&_pipelines[i]);
    }
    free(_pipelines);
    rtDestroyRWLock(&_lock);
    _first_free = NULL;
}

rt_pipeline_handle RT_RENDERER_API_FN(CompilePipeline)(const rt_pipeline_info *info) {
    rt_pipeline_handle handle = RT_INVALID_HANDLE;
    rtLockWrite(&_lock);
    if (!_first_free) {
        rtLog("VK", "No free pipeline slots!");
        rtUnlockWrite(&_lock);
        return handle;
    }
    rt_pipeline_slot *slot = _first_free;
    _first_free            = slot->next_free;
    slot->version = (slot->version + 1) & RT_GFX_HANDLE_MAX_VERSION;
    rtUnlockWrite(&_lock);

    /* No other thread that calls compile gets the same slot.
     * Another thread accessing the slot via GetPipeline would get a version mismatch.
     * The same holds for DestroyPipeline
     */

    return handle;
}

void RT_RENDERER_API_FN(DestroyPipeline)(rt_pipeline_handle handle) {
    if (handle.index >= (uint32_t)r_VkMaxPipelineCount.i)
        return;
    rtLockWrite(&_lock);
    if (_pipelines[handle.index].version == handle.version)
        DestroyPipeline(&_pipelines[handle.index]);
    else
        rtLog("VK", "Tried to destroy a pipeline using an outdated handle.");
    rtUnlockWrite(&_lock);
}

const rt_pipeline *rtGetPipeline(rt_pipeline_handle handle) {
    if (handle.index >= (uint32_t)r_VkMaxPipelineCount.i)
        return NULL;
    rtLockRead(&_lock);
    rt_pipeline *res = NULL;
    if (_pipelines[handle.index].version == handle.version)
        res = &_pipelines[handle.index].pipeline;
    else
        rtLog("VK", "Tried to access a pipeline using an outdated handle.");
    rtUnlockRead(&_lock);
    return res;
}
