#include "runtime.h"
#include "mem_arena.h"
#include "description_parser.h"
#include "buffer_manager.h"
#include "asset_compiler.h"
#include "shader_compiler.h"
#include "gfx.h"
#include "config.h"

#include <limits.h>
#include <assert.h>
#include <string.h>

typedef struct {
    rt_attribute_binding *uniform_bindings;
    rt_attribute_binding *storage_bindings;
    rt_attribute_binding *texture_bindings;

    rt_shader_bytecode vertex_shader;
    rt_shader_bytecode fragment_shader;
    rt_shader_bytecode compute_shader;

    /* TODO Fixed function settings */

    /* Sampler settings */

    uint16_t uniform_binding_count;
    uint16_t storage_binding_count;
    uint16_t texture_binding_count;
} rt_parsed_pipeline_data;

enum {
    RT_SHADER_NOT_PRESENT = RT_ASSET_PROCESSING_FAILED + 1
};

extern rt_cvar rt_Renderer;

#if 0
static void
DbgPrintShaderFile(const rt_parse_state *state, unsigned int list_index, unsigned int indent) {
    assert(list_index < state->statement_list_count);
    const rt_parsed_stmt_list *list = &state->statement_lists[list_index];

    unsigned int stmt_index = list->first;
    for (unsigned int i = 0; i < list->count; ++i) {
        const rt_parsed_stmt *stmt = &state->statements[stmt_index];
        for (unsigned int j = 0; j < indent; ++j)
            printf(" ");
        printf("%.*s: ", stmt->attribute.length, stmt->attribute.start);
        if (stmt->form == RT_STMT_FORM_VALUE) {
            printf("%.*s\n", stmt->value.length, stmt->value.start);
        } else {
            printf("{\n");
            DbgPrintShaderFile(state, stmt->list_index, indent + 2);
            printf("}\n");
        }
        stmt_index = stmt->next;
    }
    assert(stmt_index == UINT_MAX || stmt_index == 0);
}
#endif

static bool ParseBindingIndex(rt_text_span span, unsigned int *index) {
    if (span.length == 0)
        return false;
    int at           = (int)span.length - 1;
    unsigned int exp = 1;
    unsigned int n   = 0;
    while (at >= 0) {
        if (span.start[at] >= '0' && span.start[at] <= '9') {
            unsigned int digit = (unsigned int)(span.start[at] - '0');
            n += digit * exp;
        } else {
            rtReportError("GFX", "Unexpected non-digit character in binding index");
            return false;
        }
        --at;
        exp *= 10;
    }
    *index = n;
    return true;
}

static rt_attribute_value ParseBindingValue(rt_text_span span) {
    if (rtCompareSpanToString(span, "MATERIAL_ALBEDO") == 0) {
        return RT_ATTRIBUTE_VALUE_MATERIAL_ALBEDO;
    } else if (rtCompareSpanToString(span, "MATERIAL_NORMAL") == 0) {
        return RT_ATTRIBUTE_VALUE_MATERIAL_NORMAL;
    }
    rtReportError("GFX", "Unsupported binding value %*.s", span.length, span.start);
    return RT_ATTRIBUTE_VALUE_UNDEFINED;
}

static bool ParseBindings(rt_parse_state *state,
                          unsigned int root_list,
                          const char *name,
                          const char *file_path,
                          rt_attribute_binding **p_bindings,
                          uint16_t *p_binding_count) {
    const rt_parsed_stmt *bindings = rtFindStatement(state, root_list, name);
    if (bindings) {
        if (bindings->form != RT_STMT_FORM_LIST) {
            rtReportError("GFX",
                          "Expected list of bindings as the value of "
                          "\"%s\" in %s",
                          name,
                          file_path);
            return false;
        }
        const rt_parsed_stmt_list *binding_list = &state->statement_lists[bindings->list_index];
        rt_attribute_binding *shader_bindings =
            rtAllocBuffer(sizeof(rt_attribute_binding) * binding_list->count);
        if (!bindings) {
            rtReportError("GFX", "Out of memory");
            return false;
        }
        unsigned int binding_count = binding_list->count;

        unsigned int stmt_index = binding_list->first;
        for (unsigned int i = 0; i < binding_list->count; ++i) {
            const rt_parsed_stmt *stmt = &state->statements[stmt_index];
            if (!ParseBindingIndex(stmt->attribute, &shader_bindings[i].index)) {
                rtReleaseBuffer(shader_bindings,
                                sizeof(rt_attribute_binding) * binding_list->count);
                return false;
            }

            shader_bindings[i].value = ParseBindingValue(stmt->value);
            if (shader_bindings[i].value == RT_ATTRIBUTE_VALUE_UNDEFINED) {
                rtReleaseBuffer(shader_bindings,
                                sizeof(rt_attribute_binding) * binding_list->count);
                return false;
            }
            stmt_index = stmt->next;
        }

        *p_bindings      = shader_bindings;
        *p_binding_count = (uint16_t)binding_count;
        return true;
    } else {
        *p_bindings      = NULL;
        *p_binding_count = 0;
        return true;
    }
}

static rt_result ParseShader(rt_parse_state *state,
                             unsigned int root_list,
                             const char *name,
                             const char *file_path,
                             rt_shader_type type,
                             rt_shader_stage stage,
                             rt_shader_optimization_level optimization,
                             rt_shader_bytecode *p_shader_bytecode,
                             rt_arena *arena) {
    const rt_parsed_stmt *stmt = rtFindStatement(state, root_list, name);
    if (stmt) {
        if (stmt->form != RT_STMT_FORM_LIST) {
            rtReportError("GFX",
                          "Expected a list as the value of "
                          "\"%s\" in %s",
                          name,
                          file_path);
            return RT_ASSET_PROCESSING_FAILED;
        }
        const rt_parsed_stmt_list *shader_list = &state->statement_lists[stmt->list_index];

        unsigned int stmt_index = shader_list->first;
        for (unsigned int i = 0; i < shader_list->count; ++i) {
            const rt_parsed_stmt *shader = &state->statements[stmt_index];
            if (shader->form != RT_STMT_FORM_VALUE && shader->form != RT_STMT_FORM_BLOCK) {
                rtReportError("GFX",
                              "Expected a simple statement or block for "
                              "\"%s.%*s\" in %s",
                              name,
                              (int)shader->attribute.length,
                              shader->attribute.start,
                              file_path);
                return RT_ASSET_PROCESSING_FAILED;
            }
            rt_shader_type in_file_type = RT_SHADER_TYPE_INVALID;
            if (rtCompareSpanToString(shader->attribute, "vk") == 0) {
                in_file_type = RT_SHADER_TYPE_VULKAN;
            } else {
                rtReportError("GFX",
                              "Invalid renderer backend"
                              "\"%*s\" in %s of file %s",
                              (int)shader->attribute.length,
                              shader->attribute.start,
                              name,
                              file_path);
                return RT_ASSET_PROCESSING_FAILED;
            }

            if (in_file_type == type) {
                if (shader->form == RT_STMT_FORM_BLOCK) {
                    /* Inline code */
                    *p_shader_bytecode =
                        CompileShader(type, stage, optimization, shader->block, file_path, arena);
                    if (!p_shader_bytecode->bytes)
                        return RT_ASSET_PROCESSING_FAILED;
                    break;
                } else if (shader->form != RT_STMT_FORM_VALUE) {
                    /* A filename */
                    rtLog("AC", "Shader source files not implemented yet!");
                    return RT_ASSET_PROCESSING_FAILED;
                }
            }

            stmt_index = shader->next;
        }
        return RT_SUCCESS;
    }
    return RT_SHADER_NOT_PRESENT;
}

static rt_shader_optimization_level
ParseOptimizationLevel(rt_parse_state *state, unsigned int root_list, const char *file_path) {
    const rt_parsed_stmt *stmt = rtFindStatement(state, root_list, "optimization");
    rt_shader_optimization_level optimization_level = RT_SHADER_OPTIMIZATION_NONE;
    if (stmt) {
        if (stmt->form != RT_STMT_FORM_VALUE) {
            rtReportError("GFX",
                          "Expected a simple statement for"
                          "\"optimization\" in %s",
                          file_path);
            return RT_SHADER_OPTIMIZATION_NONE;
        }

        if (rtCompareSpanToString(stmt->value, "speed") == 0) {
            optimization_level = RT_SHADER_OPTIMIZATION_SPEED;
        } else if (rtCompareSpanToString(stmt->value, "size") == 0) {
            optimization_level = RT_SHADER_OPTIMIZATION_SIZE;
        } else if (rtCompareSpanToString(stmt->value, "none") == 0) {
            optimization_level = RT_SHADER_OPTIMIZATION_NONE;
        } else {
            rtReportError("GFX",
                          "Expected one of 'speed', 'size' and 'none' for \"optimization\" in %s",
                          file_path);
        }
    }

    return optimization_level;
}

static rt_result ParsePipelineFile(rt_file_id fid,
                                   const char *text,
                                   size_t length,
                                   rt_parsed_pipeline_data *pipeline,
                                   rt_arena *arena) {
    /* This is the grammar for pipeline files:
     * <stmt-list> ::= <stmt>*
     * <stmt> ::= <attribute> ( ( <value> ';' ) | ( '{' <stmt-list> '}' ) )
     * <attribute> ::= [:alnum:]*
     * <value>:: = [:alnum:]* */
    const char *file_path = rtGetFilePath(fid);
    rt_parse_state state;
    unsigned int root_list;
    rt_result result = rtParseDescription(text, length, file_path, &root_list, &state, arena);
    if (result != RT_SUCCESS) {
        goto out;
    }

    /* We allow the pipeline file to overwrite the optimization level */
    rt_shader_optimization_level optimization = ParseOptimizationLevel(&state, root_list, file_path);
    
    rt_shader_type type = RT_SHADER_TYPE_INVALID;
    if (strcmp(rt_Renderer.s, "vk") == 0)
        type = RT_SHADER_TYPE_VULKAN;

    if (type == RT_SHADER_TYPE_INVALID) {
        result = RT_ASSET_PROCESSING_FAILED;
        rtLog("AC", "rt_Renderer (%s) could not be translated to a shader type.", rt_Renderer.s);
        goto out;
    }

    /* Process shader stages */
    if (ParseShader(&state,
                    root_list,
                    "vertex",
                    file_path,
                    type,
                    RT_SHADER_STAGE_VERTEX,
                    optimization,
                    &pipeline->vertex_shader,
                    arena) == RT_ASSET_PROCESSING_FAILED) {
        result = RT_ASSET_PROCESSING_FAILED;
        goto out;
    }
    if (ParseShader(&state,
                    root_list,
                    "fragment",
                    file_path,
                    type,
                    RT_SHADER_STAGE_FRAGMENT,
                    optimization,
                    &pipeline->fragment_shader,
                    arena) == RT_ASSET_PROCESSING_FAILED) {
        result = RT_ASSET_PROCESSING_FAILED;
        goto out;
    }
    if (ParseShader(&state,
                    root_list,
                    "compute",
                    file_path,
                    type,
                    RT_SHADER_STAGE_COMPUTE,
                    optimization,
                    &pipeline->compute_shader,
                    arena) == RT_ASSET_PROCESSING_FAILED) {
        result = RT_ASSET_PROCESSING_FAILED;
        goto out;
    }

    /* Process bindings */
    pipeline->texture_bindings      = NULL;
    pipeline->texture_binding_count = 0;
    pipeline->uniform_bindings      = NULL;
    pipeline->uniform_binding_count = 0;
    pipeline->storage_bindings      = NULL;
    pipeline->storage_binding_count = 0;

    if (!ParseBindings(&state,
                       root_list,
                       "texture_bindings",
                       file_path,
                       &pipeline->texture_bindings,
                       &pipeline->texture_binding_count)) {
        result = RT_ASSET_PROCESSING_FAILED;
        goto out;
    }

    if (!ParseBindings(&state,
                       root_list,
                       "uniform_bindings",
                       file_path,
                       &pipeline->uniform_bindings,
                       &pipeline->uniform_binding_count)) {
        result = RT_ASSET_PROCESSING_FAILED;
        goto out;
    }

    if (!ParseBindings(&state,
                       root_list,
                       "storage_bindings",
                       file_path,
                       &pipeline->storage_bindings,
                       &pipeline->storage_binding_count)) {
        result = RT_ASSET_PROCESSING_FAILED;
        goto out;
    }

out:
    return result;
}

rt_result PipelineProcessor(rt_file_id file, rt_arena *arena) {
    rt_loaded_asset asset = LoadAsset(file);
    if (!asset.buffer)
        return RT_UNKNOWN_ERROR;

    rt_parsed_pipeline_data pipeline;
    rt_result result = ParsePipelineFile(file, asset.buffer, asset.size, &pipeline, arena);
    if (result != RT_SUCCESS)
        goto out;

    

out:
    rtReleaseBuffer(asset.buffer, asset.size);
    return result;
}