#include "assetmeta.h"
#include "processing.h"
#include "utils.h"
#include "packages.h"
#include "dependency_tracking.h"

#define RT_ASSETC_DONT_DEFINE_OPTIONS_GLOBAL
#include "options.h"

#include "runtime/aio.h"
#include "runtime/buffer_manager.h"
#include "runtime/uidtab.h"

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

extern rt_result rtProcessPipelineFile(rt_file_id file,
                                       void *buffer,
                                       size_t size,
                                       uint32_t flags,
                                       rt_processor_output *output);

extern rt_result rtProcessShaderFile(rt_file_id file,
                                     void *buffer,
                                     size_t size,
                                     uint32_t flags,
                                     rt_processor_output *output);

extern void rtDiscoverAssets(void);

rt_assetc_options g_assetc_options = {
    .root_directory   = ".",
    .optimization     = RT_ASSET_OPTIMIZATION_NONE,
    .renderer_backend = RT_RENDERER_BACKEND_CODE_VK,
};

static bool ParseCommandLineArgs(int argc, char **argv) {
    bool have_root_directory = false;
    for (int i = 1; i < argc; ++i) {
        if (i < argc - 1) {
            const char *val = argv[i + 1];
            bool matched    = false;

            if (strcmp(argv[i], "-r") == 0 || strcmp(argv[i], "--renderer") == 0) {
                if (strcmp(val, "vk") == 0) {
                    g_assetc_options.renderer_backend = RT_RENDERER_BACKEND_CODE_VK;
                } else {
                    rtReportError("ASSETC",
                                  "Invalid render backend %s. Valid options are: vk",
                                  val);
                    return false;
                }
                matched = true;
            } else if (strcmp(argv[i], "-o") == 0 || strcmp(argv[i], "--optimization") == 0) {
                if (strcmp(val, "none") == 0) {
                    g_assetc_options.optimization = RT_ASSET_OPTIMIZATION_NONE;
                } else if (strcmp(val, "space") == 0) {
                    g_assetc_options.optimization = RT_ASSET_OPTIMIZATION_SPACE;
                } else if (strcmp(val, "performance") == 0) {
                    g_assetc_options.optimization = RT_ASSET_OPTIMIZATION_PERFORMANCE;
                } else {
                    rtReportError("ASSETC",
                                  "Invalid optimization level %s. Valid options are: none, space, "
                                  "performance",
                                  val);
                    return false;
                }
                matched = true;
            } else if (argv[i][0] == '-') {
                rtReportError("ASSETC", "Invalid command line argument %s", argv[i]);
                return false;
            }

            if (matched) {
                /* Skip the value */
                ++i;
                continue;
            }
        }

        if (!have_root_directory) {
            g_assetc_options.root_directory = argv[i];
        } else {
            /* Maybe have secondary directories later? */
            rtReportError("ASSETC",
                          "More than one root directory passed as command line "
                          "argument.");
            return false;
        }
    }
    return true;
}

int main(int argc, char **argv) {
    rtInitRuntime();

    /* Init assetc */
    if (!ParseCommandLineArgs(argc, argv)) {
        return 1;
    }
    rtInitPackages();
    

    if (rtLoadAssetMeta() != RT_SUCCESS) {
        return 1;
    }
    
    if (rtInitDependencyTracking() != RT_SUCCESS) {
        return 1;
    }

    if (rtAddAssetProcessor(".pipeline", rtProcessPipelineFile) != RT_SUCCESS)
        return 1;
    if (rtAddAssetProcessor(".glsl", rtProcessShaderFile) != RT_SUCCESS)
        return 1;

    if (rtStartProcessing() != RT_SUCCESS) {
        return 1;
    }

    /* Create necessary directories */
    rtSetWorkingDirectory(g_assetc_options.root_directory);
    rtCreateDirectory("assets");
    rtCreateDirectory("actemp");
    rtCreateDirectory("data");

    /* "Mainloop" */
    rtDiscoverAssets();

    rtWaitUntilProcessingIsFinished();
    rtStopProcessing();

    /* Write result */
    rtSaveAssetMeta();
    if (rtSavePackages() != RT_SUCCESS) {
        return 1;
    }
    if (rtWriteUIDTab() != RT_SUCCESS) {
        return 1;
    }
    if (rtSaveAssetDependencies() != RT_SUCCESS) {
        return 1;
    }

    rtShutdownRuntime();

    return 0;
}