#include "assetmeta.h"
#include "utils.h"
#include "options.h"

#include "runtime/threading.h"
#include "runtime/aio.h"
#include "runtime/buffer_manager.h"

#include <xxhash/xxhash.h>
#include <stdio.h>
#include <string.h>

#define MAP_SIZE 2048
typedef struct {
    rt_file_id fids[MAP_SIZE];
    rt_uid uids[MAP_SIZE];
    rt_assetmeta meta[MAP_SIZE];

    unsigned int used_slots;
} rt_uid_map;

#pragma pack(push, 1)
typedef struct {
    XXH64_canonical_t checksum;
    uint32_t num_entries;
    uint32_t _reserved;
} rt_assetmeta_header;
#pragma pack(pop)

#pragma pack(push, 1)
typedef struct {
    rt_uid uid;
    rt_file_id source_file;
    rt_assetmeta meta;
} rt_assetmeta_entry;
#pragma pack(pop)

static rt_uid_map _map;
static rt_mutex *_guard;

rt_result rtLoadAssetMeta(void) {
    _guard = rtCreateMutex();
    
    /* Load the meta file */
    size_t fsz = rtGetFileSize("actemp/meta.bin");
    if (fsz == 0) {
        rtLog("ASSETC", "Metadata file 'meta.bin' not found. All assets will be processed.");
        return RT_SUCCESS;
    }

    void *buffer = rtAllocBuffer(fsz);
    if (!buffer) {
        rtReportError("ASSETC", "Failed to allocate buffer for holding asset metadata.");
        return 1;
    }
    
    rt_load_batch load;
    load.loads[0].dest      = buffer;
    load.loads[0].file      = rtAddFile("actemp/meta.bin");
    load.loads[0].num_bytes = fsz;
    load.loads[0].offset    = 0;
    load.num_loads          = 1;

    rt_aio_handle handle;
    if (rtSubmitLoadBatch(&load, &handle) != RT_SUCCESS) {
        rtReportError("ASSETC", "Failed to submit load of 'meta.bin'.");
        rtReleaseBuffer(buffer, fsz);
        return 1;
    }
    if (rtWaitForAIOCompletion(handle) != RT_AIO_STATE_FINISHED) {
        rtReportError("ASSETC", "Failed to load 'meta.bin'.");
        rtReleaseBuffer(buffer, fsz);
        rtReleaseAIO(handle);
        return 1;
    }
    rtReleaseAIO(handle);

    const rt_assetmeta_header *header = buffer;
    const rt_assetmeta_entry *entries = (rt_assetmeta_entry *)(header + 1);
    
    if ((sizeof(rt_assetmeta_entry) * header->num_entries + sizeof(*header)) > fsz) {
        rtReportError("ASSETC", "Header of 'meta.bin' is corrupted: Mismatched num_entries and filesize.");
        rtReleaseBuffer(buffer, fsz);
        return 1;
    }
    XXH64_hash_t hash = XXH3_64bits(entries, sizeof(rt_assetmeta_entry) * header->num_entries);
    XXH64_hash_t header_hash = XXH64_hashFromCanonical(&header->checksum);
    if (hash != header_hash) {
        rtReportError("ASSETC",
                      "Metadata file 'meta.bin' is corrupted: Wrong checksum.");
        rtReleaseBuffer(buffer, fsz);
        return 1;
    }

    for (uint32_t i = 0; i < header->num_entries; ++i) {
        /* Load here to avoid unaligned pointer */
        rt_assetmeta meta = entries[i].meta;
        rtAddUIDMapping(entries[i].source_file, entries[i].uid, &meta);
    }

    rtReleaseBuffer(buffer, fsz);
    return 0;
}

rt_result rtSaveAssetMeta(void) {
    rt_assetmeta_header header = {0};

    /* Count number of entries */
    for (size_t i = 0; i < MAP_SIZE; ++i) {
        if (_map.fids[i] != RT_INVALID_FILE_ID)
            header.num_entries += 1;
    }

    rt_assetmeta_entry *entries = rtAllocBuffer(sizeof(rt_assetmeta_entry) * header.num_entries);
    if (!entries)
        return RT_UNKNOWN_ERROR;

    /* Store entries */
    size_t j = 0;
    for (size_t i = 0; i < MAP_SIZE; ++i) {
        if (_map.fids[i] != RT_INVALID_FILE_ID) {
            entries[j].source_file = _map.fids[i];
            entries[j].uid         = _map.uids[i];
            entries[j].meta        = _map.meta[i];
            ++j;
        }
    }

    XXH64_hash_t hash = XXH3_64bits(entries, sizeof(rt_assetmeta_entry) * header.num_entries);
    XXH64_canonicalFromHash(&header.checksum, hash);
    header._reserved = 0;

    FILE *f = fopen("actemp/meta.bin", "wb");
    if (!f) {
        rtReportError("ASSETC", "Failed to open 'meta.bin'");
        rtReleaseBuffer(entries, sizeof(rt_assetmeta_entry) * header.num_entries);
        return RT_UNKNOWN_ERROR;
    }
    if (fwrite(&header, sizeof(header), 1, f) != 1) {
        rtReportError("ASSETC", "Failed to write 'meta.bin'");
        rtReleaseBuffer(entries, sizeof(rt_assetmeta_entry) * header.num_entries);
        return RT_UNKNOWN_ERROR;
    }
    if (fwrite(entries, sizeof(rt_assetmeta_entry), header.num_entries, f) != header.num_entries) {
        rtReportError("ASSETC", "Failed to write 'meta.bin'");
        rtReleaseBuffer(entries, sizeof(rt_assetmeta_entry) * header.num_entries);
        return RT_UNKNOWN_ERROR;
    }

    fclose(f);

    return RT_SUCCESS;
}

rt_uid rtLookupUID(rt_file_id fid) {
    rtLockMutex(_guard);
    unsigned int i = 0;
    rt_uid result  = RT_INVALID_UID;
    while (i < MAP_SIZE) {
        unsigned int slot = (fid + i) % MAP_SIZE;
        if (_map.fids[slot] == fid) {
            result          = _map.uids[slot];
            break;
        } else if (_map.fids[slot] == RT_INVALID_FILE_ID) {
            break;
        }
        ++i;
    }
    rtUnlockMutex(_guard);
    return result;
}

bool rtGetAssetMeta(rt_file_id source_file, rt_assetmeta *meta) {
    rtLockMutex(_guard);
    unsigned int i = 0;
    bool result    = false;
    while (i < MAP_SIZE) {
        unsigned int slot = (source_file + i) % MAP_SIZE;
        if (_map.fids[slot] == source_file) {
            *meta  = _map.meta[slot];
            result = true;
            break;
        } else if (_map.fids[slot] == RT_INVALID_FILE_ID) {
            break;
        }
        ++i;
    }
    rtUnlockMutex(_guard);
    return result;
}

void rtAddUIDMapping(rt_file_id fid, rt_uid uid, const rt_assetmeta *meta) {
    rtLockMutex(_guard);
    float fill_rate = (float)_map.used_slots / MAP_SIZE;
    if (fill_rate >= .5f) {
        rtLog("ASSETC", "UID map is above 50% filled.");
    }
    unsigned int i = 0;
    while (i < MAP_SIZE) {
        unsigned int slot = (fid + i) % MAP_SIZE;
        if (_map.fids[slot] == RT_INVALID_FILE_ID) {
            _map.fids[slot] = fid;
            _map.uids[slot] = uid;
            if (meta) {
                _map.meta[slot] = *meta;
            } else {
                _map.meta[slot].last_changed = 0;
                _map.meta[slot].compiled_ts = 0;
                _map.meta[slot].processing_flags = 0;
            }
            ++_map.used_slots;
            break;
        } else if (_map.fids[slot] == fid) {
            break;
        }
        ++i;
    }
    if (i == MAP_SIZE) {
        rtReportError("ASSETC", "Failed to insert entry into UID map.");
    }
    rtUnlockMutex(_guard);
}